/* mp3rg2sc

mp3rg2sc is a program to convert the ReplayGain values in an mp3 file into
iTunes Sound Check values as used by iTunes (and Sonos). There are other
existing routines that do this but they all fall short. There is a perl script
called rg2sc.pl  by ??? but that only runs on *ix based systems and, in any
case, seems to corrupt other COMM tags when I tested it. There is also a
routine in mp3tag, also called rg2sc but that writes all the iTunNORM
fields to the same value ignoring some of the known information.

It supports mp3	ID3v2.2, 2.3 & 2.4

The program is a command line tool, intended to be called inline by
a media manager such as MediaMonkey. In normal operation it is silent,
ignoring without comment minor errors such as being called for non mp3
files.

One limitation of the program is that it assumes that the first frame in
the file is an ID3v2 frame. All subsequent frames are passed through
untouched.

The mechanism to remove iTunNORM frames (-r switch) just changes the name to
xTunNORM on the grounds that nothing will find it. This saves us having to
shuffle up all the remaining frames. If re-instated by this program then
it will re-use the same but if the file is given to iTunes in the mean time
then it may create another completely new frame.

If some things seem to be done a little strangely in this code then that
is because it has been hacked out of a much larger program and adapted.

Assistance was gained from Id3Gain by Dr. Lex http://www.dr-lex.be/software/id3gain.html
and rg2sc by Richard van den Berg http://www.vdberg.org/~richard/rg2sc.html

*/
#define VERSION "2.1"

/* change log
	v1.0 - initial release
	v1.1 - ?
	v1.2 - ignore case of file extension
	v1.3 - Allow for file names containing '.'
	v1.4 - switch to read iTunNORM (the first item) and write it to STDOUT
	v2.0 - Log actions, wipe out duplicate and incorrect iTunNORM frames
	v2.1 - Also wipe out incorrect xTunNORM frames
*/

/* #define UNIX	/* Uncomment this line to compile on a unix/linux system */
		/* comment it out to compile on a Windows system */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>

void usage(void)
{
	fprintf(stderr,"Usage (v%s): mp3rg2sc [-h|-?] [-a] [-r] [-x] [-v] [-l <logfile>] <filename>\n", VERSION);
	fprintf(stderr,"\th|?=Help (this note)\n");
	fprintf(stderr,"\ta=Use Album ReplayGain\n");
	fprintf(stderr,"\tr=Remove iTunes Sound Check\n");
	fprintf(stderr,"\tx=Read iTunes Sound Check\n");
	fprintf(stderr,"\tv=Verbose (debug) info\n");
}

/* mp3rg2sc Copyright (c) 2013 Rick Parsons.

This software is published Open Source Freeware. You are free to use the
programs either in full or part without charge. I would ask however that
this notice is included with any extract or onward distribution.

No charge may be made for the use of this code, however there is no
objection to including modest extracts (with acknowledgement) in otherwise
commercial software.

No warranty is given or implied by the use of the software which you do at
your own risk. No compensation can be considered regarding damage to data,
computers or any thing else arising as a result of using these programs.

*/

#define TRUE 1
#define FALSE 0
typedef short int logical;
typedef unsigned char uint8;
typedef unsigned short int uint16;
typedef unsigned long int uint32;
typedef unsigned long long int uint64;

int options(int, char **);
void splitpath(const char *, char *, char *, char *);
logical process_file(char *, FILE *, unsigned char *, int);

void mpeg_file(FILE *, unsigned char *, int);
void mpeg_frame(FILE *, unsigned char *, int *, int);
void ID3v2_tag(FILE *, unsigned char *, unsigned char *, int *, int);
int syncsafe(unsigned char *);
void syncsafe_out(FILE *, int);
int find_type(char *);
int find_type2(char *);

int vfgetm(unsigned char *, int, FILE *, unsigned char *, int, int);
void put_hex(int, unsigned char *);
void iTunGAIN_out(FILE *, double, double);
void iTunPEAK_out(FILE *, double);
void iTunGAIN(FILE *, int);
uint16 word16(unsigned char *);
uint32 word24(unsigned char *);
uint32 word32(unsigned char *);
double get_gain(unsigned char *, int *);
int strlen_enc(int, unsigned char *, int *);
void get_enc(int, int, unsigned char *, char *);
void logger(char *);
void obliterate(FILE *, int, int, int, char *);

static int verbose, indent, ulost;
static logical albumgain, unsync, remove_iTunNORM, read_iTunNORM, logging;
static char *logfile, *current_file;

int main(int argc, char *argv[])
{
	int arg;
	FILE * in;
	extern int verbose;
	extern logical unsync, read_iTunNORM;

	if (argc < 2)
	{
		/* If no arguments we call the usage routine and exit */
		usage();
		return 1;
	}

	/* handle the program options */
	arg = options(argc,argv);

	/* If there is still no file name then what do we do? */
	if (arg == 0)
	{
		usage();
		return 2;
	}

	if (verbose > 0)
		printf("mp3rg2sc v%s\n", VERSION);

	indent = -1;
	unsync = FALSE;

	for (; arg < argc; arg++)
	{
		/* opening here is a bit premature but it makes things easier.
		If only reading then don't open for write */
		if ((in = fopen(argv[arg], read_iTunNORM? "rb" : "rb+")) == (FILE *)NULL)
		{
			fprintf(stderr, "Error: Input file %s failed to open\n", argv[arg]);
			perror("Error");
			return FALSE;
		}

		if (verbose > 0)
			printf("Processing %s\n", argv[arg]);

/* ***************************************************************************** */
/*		This is where things get done					 */
		(void)process_file(argv[arg], in, (unsigned char *)NULL, 0);
/*										 */
/* ***************************************************************************** */

		fclose(in);

	}

	if (verbose > 0)
		printf("mp3rg2sc end of job\n");

	return 0;
}

/* returns the index of the first argument that is not an option; i.e.
	does not start with a dash (unix) or a slash (DOS) */
int options(int argc,char *argv[])
{
	int i,firstnonoption=0;
	extern int verbose;
	extern logical albumgain, remove_iTunNORM, read_iTunNORM;

	verbose = 0;
	albumgain = FALSE;
	remove_iTunNORM = FALSE;
	read_iTunNORM = FALSE;
	logging = FALSE;
	for (i=1; i< argc;i++)
	{
		if (
#ifndef UNIX
			argv[i][0] == '/' ||
#endif
			argv[i][0] == '-')
		{
			switch (argv[i][1])
			{
				case 'a':
				case 'A':
					albumgain = TRUE;
					break;
				case 'r':
				case 'R':
					remove_iTunNORM = TRUE;
					break;
				case 'x':
				case 'X':
					read_iTunNORM = TRUE;
					break;
				case 'v':
				case 'V':
					verbose++;
					break;
				case 'l':
				case 'L':
					logging = TRUE;
					logfile = argv[i + 1];
					i++;
					break;
				case 'h':
				case 'H':
				case '?':
					usage();
					return 0;
				default:
					fprintf(stderr,"Unknown option %s\n",argv[i]);
					break;
			}
		}
		else
		{
			firstnonoption = i;
			break;
		}
	}
	return firstnonoption;
}

/* splits up a file name into its constituant parts (Unix and Windows) */
void splitpath(const char * string, char * path, char * file, char * ext)
{
	char * cp;

	strcpy(path, string);
	*file = '\0';
	*ext = '\0';

#ifdef UNIX
	if ((cp = strrchr(path, '/')) != (char *)NULL)
#else
	if ((cp = strrchr(path, '\\')) != (char *)NULL)
#endif
	{
		*cp = '\0';
		strcpy(file, cp + 1);
	}
	else
	{
		strcpy(file, path);
		*path = '\0';
	}

	if ((cp = strrchr(file, '.')) != (char *)NULL)
	{
		*cp = '\0';
		strcpy(ext, cp + 1);
	}

	return;
}

/* check that we actually have an MP3 file (by extension) and if so process it.
	The original catered for a number of file types
input = Input file name
in = FILE control block pointer
vfile = see vfgetm for details
*/
logical process_file(char * input, FILE * in, unsigned char * vfile, int file_length)
{
	extern char *current_file;
	char fname[256];
	char path[1024];
	char ext[10];
	char * e;

	splitpath(input, path, fname, ext);
	current_file = input; /* for logger */

	/* Check for correct file type */
	e = ext; while(*e != '\0')
		*e = tolower(*e), e++;
	if (strcmp(ext, "mp3") == 0)
		mpeg_file(in, vfile, file_length);
	else
	{
		if (verbose > 0)
			printf("File type %s not supported, skipped over %s\n", ext, input);
		return FALSE;
	}

	return TRUE;
}

void logger(char *string)
{
	extern logical logging;
	extern char *logfile, *current_file;
	FILE *lg;
	time_t rawtime;
	char buffer[25];

	if (logging)
	{
		time (&rawtime);
		strftime(buffer, 25, "%Y-%m-%d %H:%M:%S", localtime(&rawtime));
		if ((lg = fopen(logfile, "a")) == (FILE *)NULL)
			return;
		fprintf(lg, "%s %s >> %s", buffer, current_file, string);
		fclose(lg);
	}
	else
		printf("%s", string);
}

/* Read the MP3 frame and check the identifier. If it is ID3 then process the tag
	The original coped with lots of different frame types.
in = FILE control block pointer
vfile = see vfgetm for details
*/
void mpeg_frame(FILE * in, unsigned char * vfile, int * file_offset_a, int file_length)
{
	unsigned char hdr[10];
	extern int verbose;

	/* Read the frame header */
	if (vfgetm(hdr, 4, in, vfile, *file_offset_a, file_length) == EOF)
		return;

	if (verbose > 0)
	{
		printf("Offset %d, Frame header ", *file_offset_a);
		put_hex(4, hdr);
		putchar('\n');
	}

	*file_offset_a += 4;

	/* Look for a frame cookie */
	if (strncmp((char *)hdr, "ID3", 3) == 0)
	{
		ID3v2_tag(in, vfile, hdr, file_offset_a, file_length);
	}
	else
	{
		printf("Error: Frame header ");
		put_hex(4, hdr);
		printf(" ignored\n");
	}
	return;
}

int syncsafe(unsigned char * buf)
{
	int l = 0, i;
	for (i = 0; i < 4; i++)
		l = l * 128 + buf[i];
	return l;
}

void syncsafe_out(FILE * f, int v)
{
	int j;
	for (j = 3; j >= 0; --j)
		fputc((v >> j * 7) & 0x7f, f);
	return;
}

uint32 word24(unsigned char * string)
{
	uint32 v = 0;
	int i;
	for (i = 0; i < 3; i++)
		v = (v << 8) + (uint32)string[i];
	return v;
}

uint32 word32(unsigned char * string)
{
	uint32 v = 0;
	int i;
	for (i = 0; i < 4; i++)
		v = (v << 8) + (uint32)string[i];
	return v;
}

/* read a block from a file.

There are two possibilities

a) We have a file which we will read to EOF
	file_length == 0
b) We have a file of which we can read up to file_length
	file_length != 0

In all cases, file_offset is how many we have read so far, but it is
	incremeted externally.
*/
int vfgetm(unsigned char * buf, int n, FILE * in, unsigned char * vfile, int file_offset, int file_length)
{
	int i, c;
	extern logical unsync;
	extern int ulost;

	/* Check that we have enough bytes left */
	if (file_length != 0 && (file_offset + n) > file_length)
		return EOF;

	for (i = 0; i < n; i++)
	{
		c = getc(in);
		if (c == EOF)
			return EOF;
		/* if this needs unsynchronising then lose the next byte*/
		if (unsync && c == 0xFF && i != n - 1)
		{
			(void)getc(in);
			ulost++;
		}
		buf[i] = c;
	}

	return n;
}

int find_type(char * tag)
{
	int i;

	if (strcmp(tag, "COMM") == 0)
		return 1;
	else if (strcmp(tag, "TXXX") == 0)
		return 2;
	else if (strcmp(tag, "RGAD") == 0)
		return 3;
	else
		return 0;
}

int find_type2(char * tag)
{
	int i;

	if (strcmp(tag, "COM") == 0)
		return 1;
	else if (strcmp(tag, "TXX") == 0)
		return 2;
	else
		return 0;
}

/* Overwrite bad iTunNORM with dummy comment. */
void obliterate(FILE * in, int floc, int flen, int nloc, char * string)
{
	int pad, i;

	fseek(in, floc, SEEK_SET);
	fputc(0, in);
	fprintf(in, "XXX");
	fputc(0, in);
	fprintf(in, "%s", string);

	/* Pad out with spaces */
	pad = flen - 5 - strlen(string);
	for (i = 0; i < pad; i++)
		fputc(' ', in);
	/* I would have thought that the file pointer would already be set to
	the end of the frame, and it does work in Unix, but Windows requires
	it to be explicitly set else it corrupts the file */
	fseek(in, nloc, SEEK_SET);
}

/* Now we are cooking - we have the right frame type so look for the tags
	The original read and interpreted all the different tag types
in = FILE control block pointer
vfile = see vfgetm for details
hdr[10] = the frame header.
*/
void ID3v2_tag(FILE * in, unsigned char * vfile, unsigned char * hdr, int * file_offset_a, int file_length)
{
	int version, tlen, mark, flen, j, enc, glen, floc, gunsync, c, term;
	union {
		long int h;
		float f;
	} peak;
	unsigned char * value;
	extern int verbose, indent;
	extern logical albumgain, unsync, remove_iTunNORM, read_iTunNORM;
	double rg_gain, rg_track = -999999.0, rg_album = -999999.0, rg_peak = 1.0;
	int iTunNORM_loc = 0, xTunNORM_loc = 0, ID3v2len_loc = *file_offset_a + 2;
	logical expand = FALSE;
	FILE * tmp;
#ifndef UNIX
	char tempfile[512];
#endif
	char gainstring[32];

	/* Read the rest of the tag  header */
	if (vfgetm(hdr + 4, 6, in, vfile, *file_offset_a, file_length) == EOF)
		return;
	*file_offset_a += 6;

	/* extract the tag length */
	tlen = syncsafe(hdr + 6);
	version = hdr[3];

	if (verbose > 0)
	{
		printf("ID3v2.%d tag", version);
		printf(", Revision: %u, Flags: ", hdr[4]);
		put_hex(1, hdr + 5);
		printf(", Length: %d, Offset: %d", tlen, ID3v2len_loc);
		putchar('\n');
	}
	/* Set the Global Un-Synchronisation flag. */
	unsync = gunsync = (hdr[5] >> 7) == 1;

	/* start counting bytes in the tag */
	mark = 0;

	if ((version >= 2) || (version <= 4)) /* revision 2, 3 or 4 */
	{
		/* Read the frames */
		while (mark < tlen)
		{
			if (verbose > 0)
				printf("Offset %d: ", *file_offset_a);

			if (version == 2)
			{
				if (vfgetm(hdr, 6, in, vfile, *file_offset_a, file_length) == EOF)
					return;
				*file_offset_a += 6;
				mark += 6;

				if (hdr[0] == (unsigned char)'\0')
					break;
				/* extract the length */
				flen = word24(hdr + 3);
				/* terminate the frame type */
				hdr[3] = (unsigned char)'\0';

				/* find out what sort of frame it is */
				j = find_type2((char *)hdr);
			}
			else
			{
				if (vfgetm(hdr, 10, in, vfile, *file_offset_a, file_length) == EOF)
					return;
				*file_offset_a += 10;
				mark += 10;

				if (hdr[0] == (unsigned char)'\0')
					break;
				/* extract the length. This is one of the few structural
					differences between v2.3 and v2.4 */
				if (version == 3)
					flen = word32(hdr + 4);
				else
					flen = syncsafe(hdr + 4);
				/* terminate the frame type */
				hdr[4] = (unsigned char)'\0';

				/* find out what sort of frame it is */
				j = find_type((char *)hdr);

				/* Set the unsynchronisation flag appropriate to this frame */
				if (version == 4)
					unsync = gunsync | ((hdr[9] & 0x02) != 0);
			}

			/* read the field value into memory */
			value = malloc(flen);
			if (vfgetm(value, flen, in, vfile, *file_offset_a, file_length) == EOF)
				return;
			floc = *file_offset_a;
			*file_offset_a += flen + ulost;
			unsync = gunsync; /* reset for the next frame */

			if (verbose > 0)
				printf("Length: %d ", flen);

			/* look at the value */
			switch (j)
			{
				case 1: /* COM(M) enc, lang, <string> 00, <string> */
					if (verbose > 0)
						printf("[COMM] ");
					enc = value[0];
					/* this assumes that all iTunNORM codes are ISO-8859-1 or ascii ! */
					glen = strlen((char *)value + 4);
					if (glen == 8 && strncmp((char *)value + 4, "iTunNORM", 8) == 0)
					{
						if (read_iTunNORM && iTunNORM_loc != 0)
						{
							printf("Duplicate iTunNORM frame ");
							break;
						}
						if (iTunNORM_loc != 0 || xTunNORM_loc != 0)
						{
							logger("Duplicate iTunNORM frame erased.\n");
							obliterate(in, floc, flen, *file_offset_a, "Duplicate iTunNORM frame erased by mp3rg2sc.");
							break;
						}
						if (flen == 104)
							/* remember where it is */
							iTunNORM_loc = floc;
						else
						{	/* Some authors forget the trailing NULL but iTunes requires it */
							if (read_iTunNORM)
								printf("Incorrect length iTunNORM frame ");
							else
							{
								logger("Incorrect length iTunNORM frame erased\n");
								obliterate(in, floc, flen, *file_offset_a, "Incorrect length iTunNORM frame erased by mp3rg2sc.");
							}
						}
						if (verbose > 0)
							printf("<<iTunNORM>> Length = %d", flen);
					}
					if (glen == 8 && strncmp((char *)value + 4, "xTunNORM", 8) == 0)
					{
						if (iTunNORM_loc != 0 || xTunNORM_loc != 0)
						{
							logger("Duplicate xTunNORM frame erased.\n");
							obliterate(in, floc, flen, *file_offset_a, "Duplicate xTunNORM frame erased by mp3rg2sc.");
							break;
						}
						if (flen == 104)
							/* remember where it is */
							xTunNORM_loc = floc;
						else if (!read_iTunNORM)
						{ /* I thought that this couldn't happen but it seems that some editors remove the trailing NULL
							from other people's comments! */
							logger("Incorrect length xTunNORM frame erased\n");
							obliterate(in, floc, flen, *file_offset_a, "Incorrect length xTunNORM frame erased by mp3rg2sc.");
						}
						if (verbose > 0)
							printf("<<xTunNORM>> Length = %d", flen);
					}
					if (verbose > 0)
						putchar('\n');
					break;

				case 2: /* TXX(X) enc, <string> 00, <string> */
					if (verbose > 0)
						printf("[TXXX] ");
					enc = value[0];
					glen = strlen_enc(enc, value + 1, &term);
					get_enc(enc, glen, value + 1, gainstring);
					if (strcmp(gainstring, "replaygain_track_gain") == 0)
					{
						get_enc(enc, flen - glen - term - 1, value + glen + term + 1, gainstring);
						sscanf(gainstring, "%8lf", &rg_track);
						if (verbose > 0)
							printf("TRACK %f\n", rg_track);
					}
					else if (strcmp(gainstring, "replaygain_album_gain") == 0)
					{
						get_enc(enc, flen - glen - term - 1, value + glen + term + 1, gainstring);
						sscanf(gainstring, "%8lf", &rg_album);
						if (verbose > 0)
							printf("ALBUM %f\n", rg_album);
					}
					else if (strcmp(gainstring, "replaygain_track_peak") == 0)
					{
						get_enc(enc, flen - glen - term - 1, value + glen + term + 1, gainstring);
						sscanf(gainstring, "%8lf", &rg_peak);
						if (verbose > 0)
							printf("PEAK %f\n", rg_peak);
					}
					break;

				case 3: /* RGAD */
					if (verbose > 0)
						printf("[RGAD] ");
					peak.h = (long int)word32(value);
					rg_peak = peak.f;
					for (glen = 4; glen < flen; glen += 2)
					{
						rg_gain = get_gain(value + glen, &enc);
						if (enc == 1)
							rg_track = rg_gain;
						else if (enc == 2)
							rg_album = rg_gain;
					}
					if (verbose > 0)
						printf("RG P%f T%f A%f\n", rg_peak, rg_track, rg_album);
					break;
				default: /* others are ignored */
					if (verbose > 0)
						printf("[other]\n");
					break;
			}
			free(value);
			mark += flen + ulost;
			ulost = 0;
		}
	}
	else
		logger("Unsupported ID3v2 tag\n");

	if (verbose > 0)
		putchar('\n');

	/* now we have got here we can collect up what we have
		and see if we have enough information */
	if (iTunNORM_loc > 0)
	{
		/* Move the pointer back to the start of the frame */
		iTunNORM_loc -= 6;
		if (version != 2)
			iTunNORM_loc -= 4;
	}
	if (xTunNORM_loc > 0)
	{
		/* Move the pointer back to the start of the frame */
		xTunNORM_loc -= 6;
		if (version != 2)
			xTunNORM_loc -= 4;
	}
	if (remove_iTunNORM)
	{
		if (iTunNORM_loc > 0)
		{
			if (verbose > 0)
				printf("The iTunNORM COMM frame will be removed from offset %d\n", iTunNORM_loc);
			iTunNORM_loc += 10;
			if (version != 2)
				iTunNORM_loc += 4;
			fseek(in, iTunNORM_loc, SEEK_SET);
			fputc('x', in);
			logger("iTunNORM frame disabled\n");
		} /* keep quiet if there is no iTunNORM */
		return; /* No more to do */
	}
	if (read_iTunNORM)
	{
		if (iTunNORM_loc > 0)
		{
			iTunNORM_loc += 20;
			if (version != 2)
				iTunNORM_loc += 4;
			fseek(in, iTunNORM_loc, SEEK_SET);
			iTunGAIN(in, 1000);
		} /* keep quiet if there is no iTunNORM */
		return;
	}

	rg_gain = rg_track;
	if (albumgain)
		rg_gain = rg_album;
	if (rg_gain <= -999999.0)
	{
		if (albumgain)
			logger("ReplayGain(Album) not found\n");
		else
			logger("ReplayGain not found\n");
		return;
	}

	/* We have three potential sitiations now.
	1) We have seen iTunNORM/xTunNORM. Easy, just overwrite it.
	2) We haven't seen iTunNORM and there is space at the end of the tag.
		So we can write it there and update the tag length.
	3) We haven't seen it and there is not enough space. This is the hard
		case. We will need to extend the tag and re-write the file.
	*/

	/* we prefer to use an existing iTunNORM space if there is one so that
	we don't end up creaing multiple tags but otherwise, use an xTunNORM space
	if there is no iTunNORM one */
	if (iTunNORM_loc == 0)
		iTunNORM_loc = xTunNORM_loc;

	if (iTunNORM_loc != 0)
	{	/* Case 1 */
		if (verbose > 0)
			printf("The iTunNORM COMM frame will be overwritten at offset %d\n", iTunNORM_loc);
		logger(albumgain ? "iTunNORM frame rewritten (album)\n" : "iTunNORM frame rewritten (track)\n");
	}
	else
	{	/* Case 2 */
		iTunNORM_loc = *file_offset_a;
		/* Move the pointer back to the start of the frame */
		iTunNORM_loc -= 6;
		if (version != 2)
			iTunNORM_loc -= 4;
		if (verbose > 0)
			printf("The iTunNORM COMM frame will be written at offset %d\n", iTunNORM_loc);
		/* The space we need is COMM header = 6 or 10, encoding = 1,
			language = 3, iTunNORM title = 8, encoding = 1,
			10 * (space + 8hex) = 90, 1 null terminator
			Total: 110 to 114 */
		flen = 110;
		if (version != 2)
			flen += 4;
		if (tlen - mark < flen)
		{	/* case 3 */
			if (verbose > 0)
				printf("\tID3v2 tag padding %d bytes\n", tlen - mark);
			/* Dispose of the old padding */
			*file_offset_a += tlen - mark;
			fseek(in, *file_offset_a, SEEK_SET);
			/* write the rest of the file to a temporary file */
#ifndef UNIX
			strcpy(tempfile, getenv("TEMP"));
			strcat(tempfile, "\\mp3rg2sc.mp3");
			if ((tmp = fopen(tempfile, "wb+")) == (FILE *)NULL)
			{
				fprintf(stderr, "Error: Temporary file %s failed to open\n", tempfile);
				perror("Error");
				return;
			}
#else
			if ((tmp = tmpfile()) == (FILE *)NULL)
			{
				fprintf(stderr, "Error: Temporary file failed to open\n");
				perror("Error");
				return;
			}
#endif
			j = 0;
			while ((c = getc(in)) != EOF)
				fputc(c, tmp), j++;
			fseek(tmp, 0, SEEK_SET);
			expand = TRUE;
		}
		logger(albumgain ? "iTunNORM frame created (album)\n" : "iTunNORM frame created (track)\n");
	}

	/* Now write the new iTunNORM tag */
	fseek(in, iTunNORM_loc, SEEK_SET);
	fprintf(in, "COM");
	if (version != 2)
		fputc('M', in);
	fputc(0, in), fputc(0, in);
	if (version != 2)
		fputc(0, in);
	fputc(104, in); /* length (it is syncsafe) */
	if (version != 2)
		fputc(0, in), fputc(0, in); /* flags */
	fputc(0, in); /* encoding */
	fprintf(in, "eng"); /* Language */
	fprintf(in, "iTunNORM");
	fputc(0, in); /* encoding */
	/* now 10 x space then 8 hex digits */
	/* This method is based on rg2sc by Richard van den Berg */
	iTunGAIN_out(in, rg_gain, 1000.0);
	iTunGAIN_out(in, rg_gain, 1000.0);
	iTunGAIN_out(in, rg_gain, 2500.0);
	iTunGAIN_out(in, rg_gain, 2500.0);
	/* Richard van den Berg sets 4,5,8,9 = 00024CA8 and 6,7 = 00007FFF */
	/* Dr. Lex uses 4,5 = 00020000 and 8,9 = 00003000 but sets Peak properly */
		fprintf(in, " 0002468A 0002468A");
	/* This method is based on Id3Gain by Dr. Lex */
	iTunPEAK_out(in, rg_peak);
	iTunPEAK_out(in, rg_peak);
		fprintf(in, " 0002468A 0002468A");
	fputc(0, in); /* terminating null (required by iTunes) */

	if (expand)
	{
		if (verbose > 0)
			printf("The tag will be extended to %d bytes\n", tlen + 256 - 10);
		/* write the new padding zeros */
		for (j = 0; j < tlen + 256 - mark - flen; j++)
			fputc(0, in);
		/* Now copy the rest of the file from TMP */
		j = 0;
		while ((c = getc(tmp)) != EOF)
			fputc(c, in), j++;
		fclose(tmp);
#ifndef UNIX
		remove(tempfile);
#endif
		/* Finally go back and write the new tag length */
		fseek(in, ID3v2len_loc, SEEK_SET);
		syncsafe_out(in, tlen + 256 - 10); /* doesn't include header */
	}

	return;
}

void iTunGAIN_out(FILE * f, double gain, double base)
{
	unsigned int n;
	n = (unsigned int)round(base * pow(10.0, -gain / 10.0));
	if (n > 65535)
		n = 65535; /* 0xFFFF */
	fprintf(f, " %08x", n);
	return;
}

void iTunPEAK_out(FILE * f, double peak)
{
	unsigned int n;
	n = (unsigned int)round(peak * 32768.0);
	if (n > 32768)
		n = 32768; /* 0x8000 */
	fprintf(f, " %08x", n);
	return;
}

void iTunGAIN(FILE * f, int factor)
{
	int g;
	fscanf(f, "%8x", &g);
	printf("%.4f", -10.0 * log10((double)g / (double)factor));
	return;
}

/* Now we know we have an MP3 file, read the first frame (which must be ID3v2) and
	interpret it
	The original called mpeg_frame in a while loop but this one only interprets
	the first frame.
in = FILE control block pointer
vfile = see vfgetm for details
*/
void mpeg_file(FILE * in, unsigned char * vfile, int file_length)
{
	int i, file_offset;

	file_offset = 0;
	mpeg_frame(in, vfile, &file_offset, file_length);

	return;
}

void put_hex(int length, unsigned char * string)
{
	int j;

	printf("<0x");
	for (j = 0; j < length; j++)
		printf("%2.2x", string[j]);
	putchar('>');

	return;
}

uint16 word16(unsigned char * string)
{
	return ((uint16)string[0] << 8) + (uint16)string[1];
}

double get_gain(unsigned char * value, int * type)
{
	int v, p;
	v = word16(value);
	*type = v >> 13;
	p = (v >> 9) & 1;
	v = v & 0777;
	if (p == 0)
		return ((double)v / 10.0);
	else
		return ((double)-v / 10.0);
}

int strlen_enc(int enc, unsigned char * string, int * term)
{
	int j;

	switch (enc)
/* Encoding	0=ISO-8859-1		00
		1=UTF-16 with BOM	0000
		2=UTF-16 without BOM	0000
		3=UTF-8			00
*/
	{
		case 0:
		case 3:
		default:
			*term = 1;
			return strlen((char *)string);
			break;
		case 1:
		case 2:
			*term = 2;
			for (j = 0; ; j++)
				if (string[j] != '\0')
				{
					j++;
					continue;
				}
				else if (string[j + 1] == '\0')
					break;
			return j;
			break;
	}
}

void get_enc(int enc, int length, unsigned char * string, char * out)
{
	int j, l, m = 0;
	uint16 c;
	char * p = out;

	if (length > 0)
	{
	/* we supress a terminating null if there is one */
	l = length;
	switch (enc)
/* Encoding	0=ISO-8859-1
		1=UTF-16 with BOM
		2=UTF-16 without BOM
		3=UTF-8
*/
	{
		case 0:
		case 3:
			if (string[l - 1] == '\0')
				--l;
			for (j = 0; j < l; j++)
				*p++ = string[j];
			break;
		case 1:
			c = word16(string);
			m = 2;
			if (c == 0xfeff)
				swab((char *)string, (char *)string, l);
			else if (c == 0xfffe)
				;
			else	/* BOM not found, continue as enc 2 */
				m = 0;
		case 2:
			if ((string[l - 1] == '\0') && (string[l - 2] == '\0'))
				l -= 2;
			swab((char *)string, (char *)string, l);
			for (j = m; j < l; j += 2)
			{
				c = word16(string + j);
				*p++ = c;
			}
			break;
	}
	}
	*p = 0;

	return;
}
